@@ -4,6 +4,7 @@ IR Construction Tests
 
 from __future__ import print_function, absolute_import
 
+import sys
 import copy
 import itertools
 import pickle
@@ -24,6 +25,9 @@ flt = ir.FloatType()
 dbl = ir.DoubleType()
 
 
+PY2 = sys.version_info[0] == 2
+
+
 class TestBase(TestCase):
     """
     Utilities for IR tests.
@@ -129,11 +133,14 @@ class TestFunction(TestBase):
         # Now with parameter attributes
         func = self.function()
         func.args[0].add_attribute("zeroext")
+        func.args[1].attributes.dereferenceable = 5
+        func.args[1].attributes.dereferenceable_or_null = 10
+        func.args[3].attributes.align = 4
         func.args[3].add_attribute("nonnull")
         func.return_value.add_attribute("noalias")
         asm = self.descr(func).strip()
         self.assertEqual(asm,
-            """declare noalias i32 @"my_func"(i32 zeroext %".1", i32 %".2", double %".3", i32* nonnull %".4")"""
+            """declare noalias i32 @"my_func"(i32 zeroext %".1", i32 dereferenceable(5) dereferenceable_or_null(10) %".2", double %".3", i32* nonnull align 4 %".4")"""
             )
 
     def test_function_metadata(self):
@@ -281,36 +288,35 @@ class TestIR(TestBase):
             # None as `null`
             "types":           mod.add_metadata([None]),
             })
+        di_compileunit = mod.add_debug_info("DICompileUnit", {
+            "language":        ir.DIToken("DW_LANG_Python"),
+            "file":            di_file,
+            "producer":        "ARTIQ",
+            "runtimeVersion":  0,
+            "isOptimized":     True,
+        }, is_distinct=True)
         di_func = mod.add_debug_info("DISubprogram", {
             "name":            "my_func",
             "file":            di_file,
             "line":            11,
             "type":            di_func_type,
             "isLocal":         False,
+            "unit":            di_compileunit,
             }, is_distinct=True)
-        di_compileunit = mod.add_debug_info("DICompileUnit", {
-            "language":        ir.DIToken("DW_LANG_Python"),
-            "file":            di_file,
-            "producer":        "ARTIQ",
-            "runtimeVersion":  0,
-            # Passing unnamed metadata as a sequence
-            "subprograms":     [di_func],
-            "isOptimized":     True,
-        }, is_distinct=True)
+
         # Check output
         strmod = str(mod)
         self.assert_ir_line('!0 = !DIFile(directory: "bar", filename: "foo")',
                             strmod)
         self.assert_ir_line('!1 = !{ null }', strmod)
         self.assert_ir_line('!2 = !DISubroutineType(types: !1)', strmod)
-        self.assert_ir_line('!3 = distinct !DISubprogram(file: !0, isLocal: false, '
-                            'line: 11, name: "my_func", type: !2)',
-                            strmod)
-        self.assert_ir_line('!4 = !{ !3 }', strmod)
-        self.assert_ir_line('!5 = distinct !DICompileUnit(file: !0, '
+        # self.assert_ir_line('!4 = !{ !3 }', strmod)
+        self.assert_ir_line('!3 = distinct !DICompileUnit(file: !0, '
                             'isOptimized: true, language: DW_LANG_Python, '
-                            'producer: "ARTIQ", runtimeVersion: 0, '
-                            'subprograms: !4)',
+                            'producer: "ARTIQ", runtimeVersion: 0)',
+                            strmod)
+        self.assert_ir_line('!4 = distinct !DISubprogram(file: !0, isLocal: false, '
+                            'line: 11, name: "my_func", type: !2, unit: !3)',
                             strmod)
         self.assert_valid_ir(mod)
 
@@ -345,6 +351,16 @@ class TestIR(TestBase):
                             strmod)
         self.assert_valid_ir(mod)
 
+    @unittest.skipUnless(PY2, 'py2 only')
+    def test_debug_info_py2_long(self):
+        mod = self.module()
+        di = mod.add_debug_info("DIBasicType",
+                                {"name": "foo",
+                                 "size": long(123)})  # long integer here
+        self.assert_ir_line('!0 = !DIBasicType(name: "foo", size: 123)',
+                            str(di))
+        self.assert_valid_ir(mod)
+
     def test_inline_assembly(self):
         mod = self.module()
         foo = ir.Function(mod, ir.FunctionType(ir.VoidType(), []), 'foo')
@@ -357,6 +373,38 @@ class TestIR(TestBase):
         self.assertInText(pat, str(mod))
         self.assert_valid_ir(mod)
 
+    def test_builder_asm(self):
+        mod = self.module()
+        foo = ir.Function(mod, ir.FunctionType(ir.VoidType(), []), 'foo')
+        builder = ir.IRBuilder(foo.append_basic_block(''))
+        asmty = ir.FunctionType(int32, [int32])
+        builder.asm(asmty, "mov $1, $2", "=r,r", [int32(123)], side_effect=True)
+        builder.ret_void()
+        pat = 'call i32 asm sideeffect "mov $1, $2", "=r,r" ( i32 123 )'
+        self.assertInText(pat, str(mod))
+        self.assert_valid_ir(mod)
+
+    def test_builder_load_reg(self):
+        mod = self.module()
+        foo = ir.Function(mod, ir.FunctionType(ir.VoidType(), []), 'foo')
+        builder = ir.IRBuilder(foo.append_basic_block(''))
+        asmty = ir.FunctionType(int32, [int32])
+        builder.load_reg(ir.IntType(64), "rax")
+        builder.ret_void()
+        pat = 'call i64 asm "", "={rax}"'
+        self.assertInText(pat, str(mod))
+        self.assert_valid_ir(mod)
+
+    def test_builder_store_reg(self):
+        mod = self.module()
+        foo = ir.Function(mod, ir.FunctionType(ir.VoidType(), []), 'foo')
+        builder = ir.IRBuilder(foo.append_basic_block(''))
+        asmty = ir.FunctionType(ir.VoidType(), [int32])
+        builder.store_reg(int64(123), ir.IntType(64), "rax")
+        builder.ret_void()
+        pat = 'call void asm sideeffect "", "{rax}" ( i64 123 )'
+        self.assertInText(pat, str(mod))
+        self.assert_valid_ir(mod)
 
 class TestGlobalValues(TestBase):
 
@@ -367,7 +415,8 @@ class TestGlobalValues(TestBase):
         globdouble = ir.GlobalVariable(mod, ir.DoubleType(), 'globdouble')
         self.assertEqual(mod.get_global('foo'), foo)
         self.assertEqual(mod.get_global('globdouble'), globdouble)
-        self.assertIsNone(mod.get_global('kkk'))
+        with self.assertRaises(KeyError):
+            mod.get_global('kkk')
         # Globals should have a useful repr()
         self.assertEqual(repr(globdouble),
                          "<ir.GlobalVariable 'globdouble' of type 'double*'>")
@@ -540,6 +589,20 @@ class TestBuildInstructions(TestBase):
                 %"d" = sub nuw nsw i32 %".1", %".2"
             """)
 
+    def test_binop_fastmath_flags(self):
+        block = self.block(name='my_block')
+        builder = ir.IRBuilder(block)
+        a, b = builder.function.args[:2]
+        # As tuple
+        builder.fadd(a, b, 'c', flags=('fast',))
+        # and as list
+        builder.fsub(a, b, 'd', flags=['ninf', 'nsz'])
+        self.check_block(block, """\
+            my_block:
+                %"c" = fadd fast i32 %".1", %".2"
+                %"d" = fsub ninf nsz i32 %".1", %".2"
+            """)
+
     def test_binops_with_overflow(self):
         block = self.block(name='my_block')
         builder = ir.IRBuilder(block)
@@ -552,12 +615,12 @@ class TestBuildInstructions(TestBase):
         builder.usub_with_overflow(a, b, 'h')
         self.check_block(block, """\
             my_block:
-                %"c" = call {i32, i1} (i32, i32) @"llvm.sadd.with.overflow.i32"(i32 %".1", i32 %".2")
-                %"d" = call {i32, i1} (i32, i32) @"llvm.smul.with.overflow.i32"(i32 %".1", i32 %".2")
-                %"e" = call {i32, i1} (i32, i32) @"llvm.ssub.with.overflow.i32"(i32 %".1", i32 %".2")
-                %"f" = call {i32, i1} (i32, i32) @"llvm.uadd.with.overflow.i32"(i32 %".1", i32 %".2")
-                %"g" = call {i32, i1} (i32, i32) @"llvm.umul.with.overflow.i32"(i32 %".1", i32 %".2")
-                %"h" = call {i32, i1} (i32, i32) @"llvm.usub.with.overflow.i32"(i32 %".1", i32 %".2")
+                %"c" = call {i32, i1} @"llvm.sadd.with.overflow.i32"(i32 %".1", i32 %".2")
+                %"d" = call {i32, i1} @"llvm.smul.with.overflow.i32"(i32 %".1", i32 %".2")
+                %"e" = call {i32, i1} @"llvm.ssub.with.overflow.i32"(i32 %".1", i32 %".2")
+                %"f" = call {i32, i1} @"llvm.uadd.with.overflow.i32"(i32 %".1", i32 %".2")
+                %"g" = call {i32, i1} @"llvm.umul.with.overflow.i32"(i32 %".1", i32 %".2")
+                %"h" = call {i32, i1} @"llvm.usub.with.overflow.i32"(i32 %".1", i32 %".2")
             """)
 
     def test_unary_ops(self):
@@ -691,6 +754,7 @@ class TestBuildInstructions(TestBase):
         c = builder.alloca(int32, name='c')
         d = builder.alloca(int32, size=42, name='d')
         e = builder.alloca(dbl, size=a, name='e')
+        e.align = 8
         self.assertEqual(e.type, ir.PointerType(dbl))
         ee = builder.store(z, e)
         self.assertEqual(ee.type, ir.VoidType())
@@ -717,7 +781,7 @@ class TestBuildInstructions(TestBase):
             my_block:
                 %"c" = alloca i32
                 %"d" = alloca i32, i32 42
-                %"e" = alloca double, i32 %".1"
+                %"e" = alloca double, i32 %".1", align 8
                 store double %".3", double* %"e"
                 store i32 %".2", i32* %"c"
                 %"g" = load i32, i32* %"c"
@@ -739,6 +803,61 @@ class TestBuildInstructions(TestBase):
             """)
         # XXX test with more complex types
 
+    def test_gep_castinstr(self):
+        # similar to:
+        # numba::runtime::nrtdynmod.py_define_nrt_meminfo_data()
+        block = self.block(name='my_block')
+        builder = ir.IRBuilder(block)
+        a, b = builder.function.args[:2]
+        int8ptr = int8.as_pointer()
+        ls = ir.LiteralStructType([int64, int8ptr, int8ptr, int8ptr, int64])
+        d = builder.bitcast(a, ls.as_pointer(), name='d')
+        e = builder.gep(d, [ir.Constant(int32, x) for x in [0, 3]], name='e')
+        self.assertEqual(e.type, ir.PointerType(int8ptr))
+        self.check_block(block, """\
+            my_block:
+                %"d" = bitcast i32 %".1" to {i64, i8*, i8*, i8*, i64}*
+                %"e" = getelementptr {i64, i8*, i8*, i8*, i64}, {i64, i8*, i8*, i8*, i64}* %"d", i32 0, i32 3
+            """)
+
+    def test_gep_castinstr_addrspace(self):
+        # similar to:
+        # numba::runtime::nrtdynmod.py_define_nrt_meminfo_data()
+        block = self.block(name='my_block')
+        builder = ir.IRBuilder(block)
+        a, b = builder.function.args[:2]
+        addrspace = 4
+        int8ptr = int8.as_pointer()
+        ls = ir.LiteralStructType([int64, int8ptr, int8ptr, int8ptr, int64])
+        d = builder.bitcast(a, ls.as_pointer(addrspace=addrspace), name='d')
+        e = builder.gep(d, [ir.Constant(int32, x) for x in [0, 3]], name='e')
+        self.assertEqual(e.type.addrspace, addrspace)
+        self.assertEqual(e.type, ir.PointerType(int8ptr, addrspace=addrspace))
+        self.check_block(block, """\
+            my_block:
+                %"d" = bitcast i32 %".1" to {i64, i8*, i8*, i8*, i64} addrspace(4)*
+                %"e" = getelementptr {i64, i8*, i8*, i8*, i64}, {i64, i8*, i8*, i8*, i64} addrspace(4)* %"d", i32 0, i32 3
+            """)
+
+    def test_gep_addrspace(self):
+        block = self.block(name='my_block')
+        builder = ir.IRBuilder(block)
+        a, b = builder.function.args[:2]
+        addrspace = 4
+        c = builder.alloca(ir.PointerType(int32, addrspace=addrspace), name='c')
+        self.assertEqual(str(c.type), 'i32 addrspace(4)**')
+        self.assertEqual(c.type.pointee.addrspace, addrspace)
+        d = builder.gep(c, [ir.Constant(int32, 5), a], name='d')
+        self.assertEqual(d.type.addrspace, addrspace)
+        e = builder.gep(d, [ir.Constant(int32, 10)], name='e')
+        self.assertEqual(e.type.addrspace, addrspace)
+        self.check_block(block, """\
+            my_block:
+                %"c" = alloca i32 addrspace(4)*
+                %"d" = getelementptr i32 addrspace(4)*, i32 addrspace(4)** %"c", i32 5, i32 %".1"
+                %"e" = getelementptr i32, i32 addrspace(4)* %"d", i32 10
+            """)
+
     def test_extract_insert_value(self):
         block = self.block(name='my_block')
         builder = ir.IRBuilder(block)
@@ -971,12 +1090,16 @@ class TestBuildInstructions(TestBase):
         builder.call(f, (a, b), 'res_f_fast', cconv='fastcc')
         res_f_readonly = builder.call(f, (a, b), 'res_f_readonly')
         res_f_readonly.attributes.add('readonly')
+        builder.call(f, (a, b), 'res_fast', fastmath='fast')
+        builder.call(f, (a, b), 'res_nnan_ninf', fastmath=('nnan', 'ninf'))
         self.check_block(block, """\
             my_block:
-                %"res_f" = call float (i32, i32) @"f"(i32 %".1", i32 %".2")
+                %"res_f" = call float @"f"(i32 %".1", i32 %".2")
                 %"res_g" = call double (i32, ...) @"g"(i32 %".2", i32 %".1")
-                %"res_f_fast" = call fastcc float (i32, i32) @"f"(i32 %".1", i32 %".2")
-                %"res_f_readonly" = call float (i32, i32) @"f"(i32 %".1", i32 %".2") readonly
+                %"res_f_fast" = call fastcc float @"f"(i32 %".1", i32 %".2")
+                %"res_f_readonly" = call float @"f"(i32 %".1", i32 %".2") readonly
+                %"res_fast" = call fast float @"f"(i32 %".1", i32 %".2")
+                %"res_nnan_ninf" = call ninf nnan float @"f"(i32 %".1", i32 %".2")
             """)
 
     def test_call_metadata(self):
@@ -993,7 +1116,7 @@ class TestBuildInstructions(TestBase):
         self.check_block(block, """\
             my_block:
                 %"a" = alloca i32
-                call void (metadata, metadata, metadata) @"llvm.dbg.declare"(metadata i32* %"a", metadata !0, metadata !0)
+                call void @"llvm.dbg.declare"(metadata i32* %"a", metadata !0, metadata !0)
             """)
 
     def test_invoke(self):
@@ -1007,7 +1130,7 @@ class TestBuildInstructions(TestBase):
         builder.invoke(f, (a, b), bb_normal, bb_unwind, 'res_f')
         self.check_block(block, """\
             my_block:
-                %"res_f" = invoke float (i32, i32) @"f"(i32 %".1", i32 %".2")
+                %"res_f" = invoke float @"f"(i32 %".1", i32 %".2")
                     to label %"normal" unwind label %"unwind"
             """)
 
@@ -1038,7 +1161,7 @@ class TestBuildInstructions(TestBase):
         self.check_block(block, """\
             my_block:
                 %"c" = icmp sgt i32 %".1", %".2"
-                call void (i1) @"llvm.assume"(i1 %"c")
+                call void @"llvm.assume"(i1 %"c")
             """)
 
 
@@ -1370,6 +1493,10 @@ class TestTypes(TestBase):
         self.assertEqual(str(ir.LiteralStructType((
             ir.PointerType(int1), ir.LiteralStructType((int32, int8))))),
             '{i1*, {i32, i8}}')
+        self.assertEqual(str(ir.LiteralStructType((int1,), packed=True)),
+                         '<{i1}>')
+        self.assertEqual(str(ir.LiteralStructType((int1,flt), packed=True)),
+                         '<{i1, float}>')
 
         # Avoid polluting the namespace
         context = ir.Context()
@@ -1380,6 +1507,10 @@ class TestTypes(TestBase):
         mytype2 = context.get_identified_type("MyType\"")
         self.assertEqual(str(mytype2), "%\"MyType\\22\"")
 
+    def test_hash(self):
+        for typ in filter(self.has_logical_equality, self.assorted_types()):
+            self.assertEqual(hash(typ), hash(copy.copy(typ)))
+
     def test_gep(self):
         def check_constant(tp, i, expected):
             actual = tp.gep(ir.Constant(int32, i))
@@ -1612,6 +1743,19 @@ class TestConstant(TestBase):
             'getelementptr ({float, i1}, {float, i1}* @"myconstant", i32 0, i32 1)')
         self.assertEqual(c.type, ir.PointerType(int1))
 
+    def test_gep_addrspace_globalvar(self):
+        m = self.module()
+        tp = ir.LiteralStructType((flt, int1))
+        addrspace = 4
+
+        gv = ir.GlobalVariable(m, tp, "myconstant", addrspace=addrspace)
+        self.assertEqual(gv.addrspace, addrspace)
+        c = gv.gep([ir.Constant(int32, x) for x in (0, 1)])
+        self.assertEqual(c.type.addrspace, addrspace)
+        self.assertEqual(str(c),
+            'getelementptr ({float, i1}, {float, i1} addrspace(4)* @"myconstant", i32 0, i32 1)')
+        self.assertEqual(c.type, ir.PointerType(int1, addrspace=addrspace))
+
     def test_bitcast(self):
         m = self.module()
         gv = ir.GlobalVariable(m, int32, "myconstant")
